#include <windows.h>
#include "displayMenu.h"
#include "menu.h"
#include "launcher.h"
#include "drawing.h"

MENU_ITEM *mallocMenuItem(int numOptions);
char *mallocString(const char *text);
MENU_ITEM *makeYesNoItem(const char *description, BOOL value);
MENU_ITEM *makeFontListItem(const char *description, char *value);
MENU_ITEM *makeFontSizeItem(const char *description, int value);
MENU_ITEM *makeNumberItem(const char *description, int value);
void freeMenuItem(MENU_ITEM *mi);
void freeMenu(MENU_ITEM **menu, int numItems);
void freeTTFontList();
void getTTFontList(HDC hdc);
int CALLBACK fontProcEx(ENUMLOGFONTEX *lpelfe, NEWTEXTMETRICEX *lpntme, int FontType, LPARAM lParam);
int compstr(const void *p1, const void *p2);
char **ttFonts;
int numTTFonts;

BOOL runConfigMenu() {
	BOOL retval;
	int numMenuItems;
	MENU_ITEM **menu;

	if (allowMenu) {
		getTTFontList(hdcMain);

		// Build the menu
		numMenuItems = 23;
		menu = (MENU_ITEM **)malloc(sizeof(MENU_ITEM *) * numMenuItems);

		menu[0] = mallocMenuItem(2);												// menuType
		menu[0]->description = mallocString("Menu Type");
		menu[0]->options[0] = mallocString("List View");
		menu[0]->options[1] = mallocString("Image Preview");
		menu[0]->selectedOption = menuType==MENUTYPE_LISTVIEW?0:1;

		menu[1] = makeFontListItem("Main Window Font", fontFace);					// fontFace
		menu[2] = makeFontSizeItem("Main Window Font Size", fontSize);				// fontSize
		menu[3] = makeFontListItem("Status Bar Font", statusBarFontFace);			// statusBarFontFace
		menu[4] = makeFontSizeItem("Status Bar Font Size", statusBarFontSize);		// statusBarFontSize
		menu[5] = makeFontSizeItem("Menu Font Size", menuFontSize);					// menuFontSize
		menu[6] = makeYesNoItem("Display Missing Games", displayMissingRoms);		// displayMissingRoms
		menu[7] = makeYesNoItem("Display Vertical Games", displayVerticalRoms);		// displayVerticalRoms
		menu[8] = makeYesNoItem("Display Horizontal Games", displayHorizontalRoms);	// displayHorizontalRoms
		menu[9] = makeYesNoItem("Display Clones", displayClones);					// displayClones
		menu[10] = makeYesNoItem("Display NeoGeo Games", displayNeoGeo);			// displayNeoGeo
		menu[11] = makeYesNoItem("Display Non-NeoGeo Games", displayNonNeoGeo);		// displayNonNeoGeo
		menu[12] = makeYesNoItem("Display Vector Games", displayVector);				// displayVector
		menu[13] = makeYesNoItem("Display Raster Games", displayRaster);			// displayRaster
		menu[14] = makeNumberItem("Display Minimum Players", minPlayers);			// minPlayers
		menu[15] = makeNumberItem("Display Maximum Players", maxPlayers);			// maxPlayers
		menu[16] = makeNumberItem("Display Minimum Buttons", minButtons);			// minButtons
		menu[17] = makeNumberItem("Display Minimum Buttons", maxButtons);			// maxButtons
		menu[18] = makeNumberItem("Display Minimum Coin Slots", minCoins);			// minCoins
		menu[19] = makeNumberItem("Display Minimum Coin Slots", maxCoins);			// maxCoins
		menu[20] = makeYesNoItem("Show Status Bar", showStatusBar);					// showStatusBar
		menu[21] = mallocMenuItem(3);												// menuType
		menu[21]->description = mallocString("Preview Pictures");
		menu[21]->options[0] = mallocString("Screen Shots");
		menu[21]->options[1] = mallocString("Flyers");
		menu[21]->options[2] = mallocString("Marquees");
		switch (previewPictures) {
			case PICTURETYPE_SCREENSHOTS: menu[21]->selectedOption = 0; break;
			case PICTURETYPE_FLYERS: menu[21]->selectedOption = 1; break;
			case PICTURETYPE_MARQUEES: menu[21]->selectedOption = 2; break;
		}
		menu[22] = mallocMenuItem(4);												// rotation
		menu[22]->description = mallocString("Screen Orientation");
		menu[22]->options[0] = mallocString("Normal");
		menu[22]->options[1] = mallocString("Inverted");
		menu[22]->options[2] = mallocString("Rotate Right");
		menu[22]->options[3] = mallocString("Rotate Left");
		switch (originalRotation) {
			case ROTATION_NORMAL: menu[22]->selectedOption = 0; break;
			case ROTATION_INVERTED: menu[22]->selectedOption = 1; break;
			case ROTATION_ROR: menu[22]->selectedOption = 2; break;
			case ROTATION_ROL: menu[22]->selectedOption = 3; break;
		}




		// Display and process the menu
		if (displayMenu(menu, numMenuItems)) {
			menuType = (menu[0]->selectedOption==0)?								// menuType
				MENUTYPE_LISTVIEW : MENUTYPE_IMAGEBROWSEVIEW;
			strcpy(fontFace, menu[1]->options[menu[1]->selectedOption]);			// fontFace
			fontSize = menu[2]->selectedOption * 2 + 8;								// fontSize
			strcpy(statusBarFontFace, menu[3]->options[menu[3]->selectedOption]);	// statusBarFontFace
			statusBarFontSize = menu[4]->selectedOption * 2 + 8;					// statusBarFontSize
			menuFontSize = menu[5]->selectedOption * 2 + 8;							// menuFontSize
			displayMissingRoms = (menu[6]->selectedOption==0)?TRUE:FALSE;			// displayMissingRoms
			displayVerticalRoms = (menu[7]->selectedOption==0)?TRUE:FALSE;			// displayVerticalRoms
			displayHorizontalRoms = (menu[8]->selectedOption==0)?TRUE:FALSE;		// displayHorizontalRoms
			displayClones = (menu[9]->selectedOption==0)?TRUE:FALSE;				// displayClones
			displayNeoGeo = (menu[10]->selectedOption==0)?TRUE:FALSE;				// displayNeoGeo
			displayNonNeoGeo = (menu[11]->selectedOption==0)?TRUE:FALSE;			// displayNonNeoGeo
			displayVector = (menu[12]->selectedOption==0)?TRUE:FALSE;				// displayVector
			displayRaster = (menu[13]->selectedOption==0)?TRUE:FALSE;				// displayRaster
			minPlayers = menu[14]->selectedOption;									// minPlayers
			maxPlayers = menu[15]->selectedOption;									// maxPlayers
			minButtons = menu[16]->selectedOption;									// minButtons
			maxButtons = menu[17]->selectedOption;									// maxButtons
			minCoins = menu[18]->selectedOption;									// minCoins
			maxCoins = menu[19]->selectedOption;									// maxCoins
			showStatusBar = (menu[20]->selectedOption==0)?TRUE:FALSE;				// showStatusBar

			switch (menu[21]->selectedOption) {		// previewPictures
				case 0: previewPictures = PICTURETYPE_SCREENSHOTS; break;
				case 1: previewPictures = PICTURETYPE_FLYERS; break;
				case 2: previewPictures = PICTURETYPE_MARQUEES; break;
		}

			switch (menu[22]->selectedOption) {										// rotation
				case 0: originalRotation = ROTATION_NORMAL; rotation = ROTATION_NORMAL; break;
				case 1: originalRotation = ROTATION_INVERTED; rotation = ROTATION_INVERTED; break;
				case 2: originalRotation = ROTATION_ROR; rotation = ROTATION_ROR; break;
				case 3: originalRotation = ROTATION_ROL; rotation = ROTATION_ROL; break;
			}

			retval = TRUE;
		} else {
			retval = FALSE;
		}

		menuCleanup();

		freeMenu(menu, numMenuItems);

		freeTTFontList();
		return retval;
	} else
		return FALSE;
}

MENU_ITEM *makeYesNoItem(const char *description, BOOL value) {
	MENU_ITEM *mi;

	mi = mallocMenuItem(2);
	mi->description = mallocString(description);
	mi->options[0] = mallocString("Yes");
	mi->options[1] = mallocString("No");
	mi->selectedOption = value?0:1;

	return mi;
}

MENU_ITEM *makeFontListItem(const char *description, char *value) {
	MENU_ITEM *mi;
	int i;

	mi = mallocMenuItem(numTTFonts);
	mi->description = mallocString(description);
	for (i=0; i<numTTFonts; i++) {
		mi->options[i] = mallocString(ttFonts[i]);
		if (!strcmp(ttFonts[i], value))
			mi->selectedOption = i;
	}

	return mi;
}

MENU_ITEM *makeNumberItem(const char *description, int value) {
	MENU_ITEM *mi;

	mi = mallocMenuItem(9);
	mi->description = mallocString(description);
	mi->options[0] = mallocString("0");
	mi->options[1] = mallocString("1");
	mi->options[2] = mallocString("2");
	mi->options[3] = mallocString("3");
	mi->options[4] = mallocString("4");
	mi->options[5] = mallocString("5");
	mi->options[6] = mallocString("6");
	mi->options[7] = mallocString("7");
	mi->options[8] = mallocString("8");
	mi->selectedOption = MAX(MIN(value, 8), 0);

	return mi;
}

MENU_ITEM *makeFontSizeItem(const char *description, int value) {
	MENU_ITEM *mi;

	mi = mallocMenuItem(15);
	mi->description = mallocString(description);
	mi->options[0] = mallocString("8");
	mi->options[1] = mallocString("10");
	mi->options[2] = mallocString("12");
	mi->options[3] = mallocString("14");
	mi->options[4] = mallocString("16");
	mi->options[5] = mallocString("18");
	mi->options[6] = mallocString("20");
	mi->options[7] = mallocString("22");
	mi->options[8] = mallocString("24");
	mi->options[9] = mallocString("26");
	mi->options[10] = mallocString("28");
	mi->options[11] = mallocString("30");
	mi->options[12] = mallocString("32");
	mi->options[13] = mallocString("34");
	mi->options[14] = mallocString("36");
	mi->selectedOption = value/2 - 4;

	return mi;
}

MENU_ITEM *mallocMenuItem(int numOptions) {
	MENU_ITEM *mi;

	mi = (MENU_ITEM *)malloc(sizeof(MENU_ITEM));
	memset(mi, 0, sizeof(MENU_ITEM));
	mi->numOptions = numOptions;
	mi->options = (char **)malloc(sizeof(char *) * numOptions);

	return mi;
}

char *mallocString(const char *text) {
	char *str;
	if (text) {
		str = (char *)malloc(strlen(text) + 1);
		strcpy(str, text);
		return str;
	} else
		return NULL;
}

void freeMenuItem(MENU_ITEM *mi) {
	int i;

	if (mi) {
		if (mi->numSubMenuItems > 0) freeMenu(mi->subMenu, mi->numSubMenuItems);
		if (mi->description) free(mi->description);
		if (mi->options) {
			for (i=0; i<mi->numOptions; i++)
				free(mi->options[i]);
			free(mi->options);
		}
		free(mi);
		mi = NULL;
	}
}

void freeMenu(MENU_ITEM **menu, int numItems) {
	int i;
	for (i=0; i<numItems; i++)
		freeMenuItem(menu[i]);

	free(menu);
	menu = NULL;
}

void freeTTFontList() {
	int i;
	for (i=0; i<numTTFonts; i++)
		free (ttFonts[i]);
	free(ttFonts);
	ttFonts = NULL;
	numTTFonts = 0;
}

void getTTFontList(HDC hdc) {
	LOGFONT lf;

	freeTTFontList();
	lf.lfCharSet = ANSI_CHARSET;
	lf.lfFaceName[0] = 0;
	lf.lfPitchAndFamily = 0;
	EnumFontFamiliesEx(hdc, &lf, (FONTENUMPROC)fontProcEx, (LPARAM)NULL, 0);

	qsort(ttFonts, numTTFonts, sizeof(char *), compstr);
}

int compstr(const void *p1, const void *p2) {
        return strcmp(*(char * const *)p1, *(char * const *)p2);
}

int CALLBACK fontProcEx(ENUMLOGFONTEX *lpelfe, NEWTEXTMETRICEX *lpntme, int FontType, LPARAM lParam) {
	if (FontType == TRUETYPE_FONTTYPE) {
		ttFonts = (char **)realloc(ttFonts, sizeof(char *) * ++numTTFonts);
		ttFonts[numTTFonts-1] = (char *)malloc(strlen(lpelfe->elfFullName) + 1);
		strcpy(ttFonts[numTTFonts-1], lpelfe->elfFullName);
	}
	return 1;
}
