#include <windows.h>
#include "launcher.h"
#include "drawing.h"
#include "mame.h"
#include <math.h>
#include "pngdib.h"
#include <stdio.h>
#include <vfw.h>

// Draw the screen to the given hdc & hbitmap but only in the rectangle provided
BOOL DrawScreenListView(HDC hdc, HBITMAP hbitmap, RECT *rect);
BOOL DrawScreenImageView(HDC hdc, HBITMAP hbitmap, RECT *rect);
BOOL DrawScreenMarqueeSelect(HDC hdc, HBITMAP hbitmap, RECT *rect);

// Draw the status bar with info on the given game
// Returns the rectangle covered by the status bar (so other stuff can avoid it)
BOOL DrawStatusBar(HDC hdc, int selectedGameIndex, int width, int height);

// Highlight the correct game
BOOL DrawHighlightToHDC(HDC hdc, int selectedGameIndex, RECT *drawArea, int fontHeight, int numDisplayRows);

// Draw the visible list of games to the given hdc
BOOL DrawGameListToHDC(HDC hdc, int selectedGameIndex, RECT *drawArea, int fontHeight, int numDisplayRows);

// Draw the correct screenshot to the given hdc if it exists
BOOL DrawPhotoToHDC(HDC hdc, HBITMAP hbitmap, GAME *game, RECT *drawArea, char *Dir, BOOL search);

// Play the attract mode video, if available
BOOL PlayAttractVideo(HDC hdc, GAME *game, RECT *drawArea );

// Draw the correct marquee to the given hdc if it exists
BOOL DrawMarqueeToHDC(HDC hdc, HBITMAP hbitmap, GAME *game, RECT *drawArea);

// Determine if a file exists
BOOL fileExists(char *filename);

// Rotates a DIB based on rotation variable in drawing.h
void rotateDIB(PNGD_P2DINFO *p2d);

BOOL DrawScreen(HDC hdc, HBITMAP hbitmap) {
#if DEBUG
	LARGE_INTEGER start, end, freq;
#endif
	BITMAP bmi;
	HBRUSH bgBrush;
	RECT rect;

#if DEBUG
	QueryPerformanceFrequency((LARGE_INTEGER*)&freq);
	QueryPerformanceCounter((LARGE_INTEGER*)&start);
#endif

	GetObject(hbitmap, sizeof(BITMAP), &bmi);

	// clear the screen (to the background color)
	rect.left = 0;
	rect.top = 0;
	rect.right = bmi.bmWidth;
	rect.bottom = bmi.bmHeight;
	bgBrush = CreateSolidBrush((menuType==MENUTYPE_LISTVIEW)?backColor:previewBackColor);
	FillRect(hdc, &rect, bgBrush);
	DeleteObject(bgBrush);

	if (gameDisplayCount > 0) {
		// Draw the status bar
		DrawStatusBar(hdc, currentGameSelection, rect.right, rect.bottom);

		// Allow for status bar
		rect.bottom -= statusBarHeight;

		// Draw everything else
		switch (menuType) {
			case MENUTYPE_IMAGEBROWSEVIEW:
				// Draw marquee selector
				DrawScreenImageView(hdc, hbitmap, &rect);

				// Draw cabinet in quadrant 2
				rect.left = rect.right / 2;
				rect.bottom = (bmi.bmHeight - statusBarHeight) / 2;
				DrawPhotoToHDC(hdc, hbitmap, gameDisplayList[currentGameSelection], &rect, cabDir, FALSE);

				// Play AVI in quadrant 4, if available
				rect.top = (bmi.bmHeight  - statusBarHeight) / 2;
				rect.bottom = bmi.bmHeight - statusBarHeight;
				// Close current video window if open
				if(attractAVI) MCIWndDestroy(attractAVI);
				// Check for and try to play attract
				if(PlayAttractVideo(hdc, gameDisplayList[currentGameSelection], &rect ) == FALSE) {
					// Draw screenshot in quadrant 4 instead
					DrawPhotoToHDC(hdc, hbitmap, gameDisplayList[currentGameSelection], &rect, ssDir, FALSE);
				}
				break;

			case MENUTYPE_LISTVIEW:
			default:
				if(attractAVI) MCIWndDestroy(attractAVI);
				DrawScreenListView(hdc, hbitmap, &rect);
		}
	}
#if DEBUG
	QueryPerformanceCounter((LARGE_INTEGER*)&end);
	printf("Screen redraw: %.2fms\n", (end.QuadPart - start.QuadPart) * 1000.0 / freq.QuadPart);
#endif

	return TRUE;
}

BOOL PlayAttractVideo(HDC hdc, GAME *game, RECT *drawArea ) {
	#if DEBUG
		LARGE_INTEGER start, end, freq;
	#endif
	char *filename;
	BOOL exists;
	int x, y;
	float scaleFactor;
	RECT rCurrent, rDest;

	#if DEBUG
		QueryPerformanceFrequency((LARGE_INTEGER*)&freq);
		QueryPerformanceCounter((LARGE_INTEGER*)&start);
	#endif

	exists = FALSE;

	filename = (char *)malloc(strlen(attractDir) + 14);	// allow room for "\8.3"
	sprintf(filename, "%s\\%s.avi", attractDir, game->romname);
	if (fileExists(filename))
			exists = TRUE;
	else {
		if (game->cloneof) {
			sprintf(filename, "%s\\%s.avi", attractDir, game->cloneof);
			if (fileExists(filename)) exists = TRUE;
		}
	}

	if (exists == TRUE) {
		attractAVI = MCIWndCreate(hwndMain,0,
							MCIWNDF_NOERRORDLG|MCIWNDF_NOMENU|MCIWNDF_NOOPEN|MCIWNDF_NOPLAYBAR,
							filename);

		// Get current video size
		MCIWndGetDest(attractAVI, &rCurrent);

		// Don't scale unless too large
		if(rCurrent.right - rCurrent.left > drawArea->right - drawArea->left) {
			scaleFactor = (float)(drawArea->right - drawArea->left)  / (rCurrent.right - rCurrent.left);
			if ( (float)(rCurrent.bottom - rCurrent.top) * scaleFactor > drawArea->bottom - drawArea->top )
				scaleFactor = (float)(drawArea->bottom - drawArea->top) / (rCurrent.bottom - rCurrent.top);
		} else
			scaleFactor = 1;

		// Scale everything to fit.
		//scaleFactor = (float)(drawArea->right - drawArea->left)  / (rCurrent.right - rCurrent.left);
		//if ( (float)(rCurrent.bottom - rCurrent.top) * scaleFactor > (drawArea->bottom - drawArea->top) )
		//	scaleFactor = (float)(drawArea->bottom - drawArea->top) / (rCurrent.bottom - rCurrent.top);

		// Centering
		drawArea->left += (float)((drawArea->right - drawArea->left) - ((rCurrent.right - rCurrent.left)* scaleFactor)) / 2;
		drawArea->top += (float)((drawArea->bottom - drawArea->top) - ((rCurrent.bottom - rCurrent.top)* scaleFactor)) / 2;

		// Apply move and resize
		SetWindowPos(attractAVI,           // window to resize
			HWND_TOP,                          // z-order: don't care
			drawArea->left, drawArea->top,                          // position: x,y
			(float)(rCurrent.right - rCurrent.left) * scaleFactor,      // width
			(float)(rCurrent.bottom - rCurrent.top) * scaleFactor,  // height
			0);

		// Start attract video and set to repeat
		//MCIWndPlay(attractAVI);
		if (0 == MCIWndSendString(attractAVI, TEXT("play repeat"))) {
			return(TRUE);
		}
		else
			return(FALSE);
	}
	else
		return(FALSE);
}

BOOL DrawScreenListView(HDC hdc, HBITMAP hbitmap, RECT *rect) {
	int numDisplayRows;
	int fontHeight;
	char *Dir;

	// set the font and forecolor for text
	SetFont(hdc, fontFace, fontSize, rotation, FALSE);
	SetTextColor(hdc, foreColor);
	SetBkMode(hdc, TRANSPARENT);

	fontHeight = GetFontHeight(hdc);
	numDisplayRows = GetNumDisplayRows(hdc, rect->right - rect->left, rect->bottom - rect->top, 0) / 2;

	// Determine the first game index to be displayed
	while (currentGameSelection < firstGameOnScreen)
		firstGameOnScreen = MAX(0, firstGameOnScreen - numDisplayRows);

	while (currentGameSelection > firstGameOnScreen + numDisplayRows-1)
		firstGameOnScreen = MIN(gameDisplayCount - numDisplayRows, firstGameOnScreen + numDisplayRows);

	// draw the highlight
	DrawHighlightToHDC(hdc, currentGameSelection, rect, fontHeight, numDisplayRows);

	// draw the screenshot
	if(previewPictures==PICTURETYPE_CABINETS) Dir=cabDir;
	else if(previewPictures==PICTURETYPE_SCREENSHOTS) Dir=ssDir;
	else if(previewPictures==PICTURETYPE_TITLES) Dir=titleDir;
	else Dir=flyerDir;

	rect->left = rect->right / 2;
	rect->top = rect->bottom / 2;

	// draw the marquee
	DrawMarqueeToHDC(hdc, hbitmap, gameDisplayList[currentGameSelection], rect);

	// Close current video window if open
	if(attractAVI) MCIWndDestroy(attractAVI);
	// Check for and try to play attract
	if(PlayAttractVideo(hdc, gameDisplayList[currentGameSelection], rect ) == FALSE) {
		// Draw screenshot in quadrant 4 instead
		DrawPhotoToHDC(hdc, hbitmap, gameDisplayList[currentGameSelection], rect, Dir, TRUE);
	}

	// draw the visible game list
	rect->left = 0;
	DrawGameListToHDC(hdc, currentGameSelection, rect, fontHeight, numDisplayRows);

	return TRUE;
}

BOOL DrawStatusBar(HDC hdc, int selectedGameIndex, int width, int height) {
	HBRUSH brush;
	RECT rect;
	int fontHeight;
	int x, y;
	int x1, y1, m1Width;
	char message[255];
	char message1[255];	// Used for game counts
	GAME *game;
	char *sortText, *previewText;

	if (showStatusBar) {
		// set the font and forecolor for text
		SetFont(hdc, statusBarFontFace, statusBarFontSize, rotation, FALSE);
		SetTextColor(hdc, statusBarForeColor);
		SetBkColor(hdc, statusBarBackColor);
		SetBkMode(hdc, OPAQUE);

		fontHeight = GetFontHeight(hdc);
		statusBarHeight = fontHeight;

		switch(sortBy) {
			case SORTBY_YEAR:	sortText="Year";	break;
			case SORTBY_MANUFACTURER:	sortText="Manufacturer";	break;
			case SORTBY_NAME:	sortText="Title";	break;
		}

		switch(previewPictures) {
				case PICTURETYPE_CABINETS:	previewText="Cabinets"; break;
				case PICTURETYPE_SCREENSHOTS:	previewText="Screenshots"; break;
				case PICTURETYPE_FLYERS:	previewText="Flyers"; break;
				case PICTURETYPE_TITLES:	previewText="Titles"; break;
				case PICTURETYPE_MARQUEES:	previewText="Marquees"; break;
		}

		if(menuType==MENUTYPE_LISTVIEW)
			sprintf(message1, "  Sort: %s - Shot: %s  %i/%i", sortText, previewText, selectedGameIndex+1, gameDisplayCount);
		else
			sprintf(message1, "  Sort: %s  %i/%i", sortText, selectedGameIndex+1, gameDisplayCount);

		m1Width = getStringWidth(hdc, message1);

		switch (rotation) {
			case ROTATION_NORMAL:
				rect.left = 0;
				rect.top = height - fontHeight;
				rect.right = width;
				rect.bottom = height;
				x = LEFT_PAD;
				y = rect.top;
				x1 = width - m1Width - LEFT_PAD;
				y1 = rect.top;
				break;
			case ROTATION_INVERTED:
				rect.left = 0;
				rect.top = 0;
				rect.right = width;
				rect.bottom = fontHeight;
				x = width - LEFT_PAD;
				y = rect.bottom;
				x1 = m1Width + LEFT_PAD;
				y1 = rect.bottom;
				break;
			case ROTATION_ROR:
				rect.left = 0;
				rect.top = 0;
				rect.right = fontHeight;
				rect.bottom = height;
				x = fontHeight;
				y = LEFT_PAD;
				x1 = fontHeight;
				y1 = height - m1Width - LEFT_PAD;
				break;
			case ROTATION_ROL:
				rect.left = width - fontHeight;
				rect.top = 0;
				rect.right = width;
				rect.bottom = height;
				x = rect.left;
				y = height - LEFT_PAD;
				x1 = rect.left;
				y1 = m1Width + LEFT_PAD;
				break;
		}

		// Fill the background of the status bar
		brush = CreateSolidBrush(statusBarBackColor);
		FillRect(hdc, &rect, brush);
		DeleteObject(brush);

		// Draw the status bar text
		game = gameDisplayList[selectedGameIndex];

		if(menuType==MENUTYPE_LISTVIEW) {
			sprintf(message, "  %i %s - %s, %s",
				game->year,
				game->manufacturer,
				(!strcmp(game->video->screen,"raster"))?"Raster":"Vector",
				game->video->orientation==VERTICAL?"Vertical":"Horizontal"
			);
		}
		else {
			sprintf(message, "  %s / %i %s - %s, %s",
							game->description,
							game->year,
							game->manufacturer,
							(!strcmp(game->video->screen,"raster"))?"Raster":"Vector",
					game->video->orientation==VERTICAL?"Vertical":"Horizontal"
			);
		}
		TextOut(hdc, x, y, message, strlen(message));
		TextOut(hdc, x1, y1, message1, strlen(message1));

	} else {	// Status bar was not drawn
		statusBarHeight = 0;
	}

	return TRUE;
}

BOOL DrawHighlightToHDC(HDC hdc, int selectedGameIndex, RECT *drawArea, int fontHeight, int numDisplayRows) {
	RECT rect;
	HBRUSH brush;
	int i;

	i = selectedGameIndex - firstGameOnScreen;

	switch (rotation) {
		case ROTATION_NORMAL:
			rect.left = drawArea->left;
			//rect.top = drawArea->top + i * fontHeight;
			rect.top = (drawArea->bottom / 2) + (i * fontHeight);
			rect.right = drawArea->right;
			//rect.bottom = drawArea->top + (i + 1) * fontHeight;
			rect.bottom = (drawArea->bottom / 2) + ((i + 1) * fontHeight);
			break;

		/* broken by PJR 10/2003
		case ROTATION_INVERTED:
			rect.left = drawArea->left;
			rect.top = drawArea->bottom - (i + 1) * fontHeight;
			rect.right = drawArea->right;
			rect.bottom = drawArea->bottom - i * fontHeight;
			break;
		case ROTATION_ROR:
			rect.left = drawArea->right - (i + 1) * fontHeight;
			rect.top = drawArea->top;
			rect.right = drawArea->right - i * fontHeight;
			rect.bottom = drawArea->bottom;
			break;
		case ROTATION_ROL:
			rect.left = drawArea->left + i * fontHeight;
			rect.top = drawArea->top;
			rect.right = drawArea->left + (i + 1) * fontHeight;
			rect.bottom = drawArea->bottom;
			break;
		*/
	}
	brush = CreateSolidBrush(highlightBackColor);
	FillRect(hdc, &rect, brush);
	DeleteObject(brush);

	return TRUE;
}

BOOL DrawGameListToHDC(HDC hdc, int selectedGameIndex, RECT *drawArea, int fontHeight, int numDisplayRows) {
	int i;
	int x, y;
	char *description;
	char *filename;
	BOOL exists;

	for (i = 0; i < numDisplayRows && i + firstGameOnScreen < gameDisplayCount; i++) {
		description = gameDisplayList[i+firstGameOnScreen]->description;
		exists = gameDisplayList[i+firstGameOnScreen]->exists;
		switch (rotation) {
			case ROTATION_NORMAL:
				x = drawArea->left + LEFT_PAD;
				//y = drawArea->top + i * fontHeight;
				y = (drawArea->bottom / 2) + (i * fontHeight);
				break;

		/* broken by PJR 10/2003
			case ROTATION_INVERTED:
				x = drawArea->right - LEFT_PAD;
				y = drawArea->bottom - i * fontHeight;
				break;
			case ROTATION_ROR:
				x = drawArea->right - i * fontHeight;
				y = drawArea->top + LEFT_PAD;
				break;
			case ROTATION_ROL:
				x = drawArea->left + i * fontHeight;
				y = drawArea->bottom - LEFT_PAD;
				break;
			*/
		}
		/* Added by PJR to show ROMS not detected in RED */
		//if(!exists) SetTextColor(hdc, 0x0000ff);
		//else SetTextColor(hdc,foreColor);
		SetTextColor(hdc,foreColor);

		// bold the font when drawing the selected game
		if (selectedGameIndex == (firstGameOnScreen+i)) {
			SetFont(hdc, fontFace, fontSize, rotation, TRUE);
		//	SetTextColor(hdc, highlightForeColor);
		}

		TextOut(hdc, x, y, description, strlen(description));

		if (selectedGameIndex == (firstGameOnScreen+i)) {
			SetFont(hdc, fontFace, fontSize, rotation, FALSE);
			SetTextColor(hdc, foreColor);
		}
	}

	return TRUE;
}

BOOL DrawPhotoToHDC(HDC hdc, HBITMAP hbitmap, GAME *game, RECT *drawArea, char *Dir, BOOL search) {
#if DEBUG
	LARGE_INTEGER start, end, freq;
#endif
	PNGD_P2DINFO p2d;
	char *filename;
	BOOL exists;
	int x, y;
	float scaleFactor;
	int previewPicturesSearch;

#if DEBUG
	QueryPerformanceFrequency((LARGE_INTEGER*)&freq);
	QueryPerformanceCounter((LARGE_INTEGER*)&start);
#endif

	exists = FALSE;
	previewPicturesSearch = 0;

	filename = (char *)malloc(strlen(Dir) + 14);	// allow room for "\8.3"
	sprintf(filename, "%s\\%s.png", Dir, game->romname);
	if (fileExists(filename))
			exists = TRUE;
	else {
		if (game->cloneof) {
			sprintf(filename, "%s\\%s.png", Dir, game->cloneof);
			if (fileExists(filename)) exists = TRUE;
		}
	}

	/* Now try to find another preview if user choice unavailable */
	if(search==TRUE) {
		while((exists==FALSE) && (previewPicturesSearch < 4)) {

			if(previewPicturesSearch==PICTURETYPE_CABINETS) Dir=cabDir;
			else if(previewPicturesSearch==PICTURETYPE_SCREENSHOTS) Dir=ssDir;
			else if(previewPicturesSearch==PICTURETYPE_TITLES) Dir=titleDir;
			else Dir=flyerDir;

			filename = (char *)malloc(strlen(Dir) + 14);	// allow room for "\8.3"
			sprintf(filename, "%s\\%s.png", Dir, game->romname);
			if (fileExists(filename))
				exists = TRUE;
			else {
				if (game->cloneof) {
					sprintf(filename, "%s\\%s.png", Dir, game->cloneof);
					if (fileExists(filename)) exists = TRUE;
				}
			}
			previewPicturesSearch++;
		}
	}

	if (exists)
	{

		ZeroMemory((void*)&p2d,sizeof(PNGD_P2DINFO));

		p2d.structsize=sizeof(PNGD_P2DINFO);
		p2d.pngfn=filename;

		if(read_png_to_dib(&p2d))
		{
			return FALSE;
		}

		// rotate the picture if needed
		rotateDIB(&p2d);

		switch (rotation) {
			case ROTATION_NORMAL:
				//x = drawArea->right - p2d.lpdib->biWidth - IMAGE_PAD;
				//y = drawArea->bottom - p2d.lpdib->biHeight - IMAGE_PAD;
				x = drawArea->left + IMAGE_PAD;
				y = drawArea->top;
				break;

		    /* broken by PJR
			case ROTATION_INVERTED:
				x = drawArea->left + IMAGE_PAD;
				y = drawArea->top + IMAGE_PAD;
				break;
			case ROTATION_ROR:
				x = drawArea->left + IMAGE_PAD;
				y = drawArea->bottom - p2d.lpdib->biHeight - IMAGE_PAD;
				break;
			case ROTATION_ROL:
				x = drawArea->right - p2d.lpdib->biWidth - IMAGE_PAD;
				y = drawArea->top + IMAGE_PAD;
				break;
				*/
		}

		// No scaling unless too large
		//if( p2d.lpdib->biWidth > drawArea->right - drawArea->left) {
		//	scaleFactor = (float)(drawArea->right - drawArea->left - IMAGE_PAD)  / p2d.lpdib->biWidth;
		//	if ( (float)p2d.lpdib->biHeight * scaleFactor > drawArea->bottom - drawArea->top )
		//		scaleFactor = (float)(drawArea->bottom - drawArea->top) / p2d.lpdib->biHeight;
		//} else
		//	scaleFactor = 1;

		// Scale everything to fit
		scaleFactor = (float)((drawArea->right - drawArea->left) - IMAGE_PAD) / p2d.lpdib->biWidth;
		if (p2d.lpdib->biHeight * scaleFactor > (drawArea->bottom - drawArea->top) )
			scaleFactor = (float)(drawArea->bottom - drawArea->top) / p2d.lpdib->biHeight;

		// Center image
		x += ((drawArea->right - drawArea->left) - p2d.lpdib->biWidth * scaleFactor) / 2;
		y += ((drawArea->bottom - drawArea->top) - p2d.lpdib->biHeight * scaleFactor) / 2;

		StretchDIBits(hdc,x,y,
			(int)(p2d.lpdib->biWidth * scaleFactor), (int)(p2d.lpdib->biHeight * scaleFactor),
			0, 0, p2d.lpdib->biWidth, p2d.lpdib->biHeight, p2d.bits, (LPBITMAPINFO)p2d.lpdib, DIB_RGB_COLORS, SRCCOPY);

		GlobalFree(p2d.lpdib);

#if DEBUG
		QueryPerformanceCounter((LARGE_INTEGER*)&end);
		printf("Loaded image: [%s] in %.2fms\n", filename, (end.QuadPart - start.QuadPart) * 1000.0 / freq.QuadPart);
#endif

	}
	free(filename);

	return TRUE;
}

BOOL DrawMarqueeToHDC(HDC hdc, HBITMAP hbitmap, GAME *game, RECT *drawArea) {
#if DEBUG
	LARGE_INTEGER start, end, freq;
#endif
	PNGD_P2DINFO p2d;
	char *filename;
	BOOL exists = FALSE;
	int x, y;
	float scaleFactor;

#if DEBUG
	QueryPerformanceFrequency((LARGE_INTEGER*)&freq);
	QueryPerformanceCounter((LARGE_INTEGER*)&start);
#endif

	filename = (char *)malloc(strlen(marqueeDir) + 14);	// allow room for "\8.3"
	sprintf(filename, "%s\\%s.png", marqueeDir, game->romname);
	if (fileExists(filename)) exists = TRUE;

	if(exists==FALSE)
	{
		if (game->cloneof) {
			sprintf(filename, "%s\\%s.png", marqueeDir, game->cloneof);
			if (fileExists(filename)) exists = TRUE;
		}
	}

	if(exists==FALSE)
	{
			sprintf(filename, "%s\\generic.png",marqueeDir);
			if (fileExists(filename)) exists = TRUE;
	}

	if (exists==TRUE) {
		ZeroMemory((void*)&p2d,sizeof(PNGD_P2DINFO));

		p2d.structsize=sizeof(PNGD_P2DINFO);
		p2d.pngfn=filename;

		if(read_png_to_dib(&p2d)) {    /* returns 0 on success */
			return FALSE;
		}

		rotateDIB(&p2d);

		switch (rotation) {
			case ROTATION_NORMAL:
				//x = drawArea->right - p2d.lpdib->biWidth - IMAGE_PAD;
				x=0;
				//y = drawArea->top + IMAGE_PAD;
				y=0;
				break;

			/* broken by PJR
			case ROTATION_INVERTED:
				x = drawArea->left + IMAGE_PAD;
				y = drawArea->bottom - p2d.lpdib->biHeight - IMAGE_PAD;
				break;
			case ROTATION_ROR:
				x = drawArea->right - p2d.lpdib->biWidth - IMAGE_PAD;
				y = drawArea->bottom - p2d.lpdib->biHeight - IMAGE_PAD;
				break;
			case ROTATION_ROL:
				x = drawArea->left + IMAGE_PAD;
				y = drawArea->top + IMAGE_PAD;
				break;
				*/
		}

	StretchDIBits(hdc,x,y,
		 (int)drawArea->right, (int)drawArea->bottom/2-(IMAGE_PAD*2),
		 0, 0, p2d.lpdib->biWidth, p2d.lpdib->biHeight, p2d.bits, (LPBITMAPINFO)p2d.lpdib, DIB_RGB_COLORS, SRCCOPY);

		GlobalFree(p2d.lpdib);

#if DEBUG
		QueryPerformanceCounter((LARGE_INTEGER*)&end);
		printf("Loaded image: [%s] in %.2fms\n", filename, (end.QuadPart - start.QuadPart) * 1000.0 / freq.QuadPart);
#endif
	}

	free(filename);

	return TRUE;
}

BOOL SetFont(HDC hdc, char *face, int size, int rotation, BOOL bold) {
	LOGFONT font;
	HFONT hFont;
	HFONT hPrevFont;

	memset(&font, 0, sizeof(LOGFONT));
    	switch(rotation) {
	        case ROTATION_NORMAL:
			font.lfEscapement = 0;
			break;
	        case ROTATION_ROR:
			font.lfEscapement = -900;
			break;
	        case ROTATION_INVERTED:
			font.lfEscapement = -1800;
			break;
	        case ROTATION_ROL:
			font.lfEscapement = 900;
			break;
		default:
			font.lfEscapement = 0;
	}

	if (bold) {
		font.lfWeight = FW_EXTRABOLD;
		font.lfItalic = 1;
	}
	strncpy(font.lfFaceName, face, LF_FACESIZE-1);
	font.lfFaceName[strlen(face)] = 0;
	font.lfHeight = -MulDiv(size, GetDeviceCaps(hdc, LOGPIXELSY), 72);
	hFont = CreateFontIndirect(&font);

	if (hFont != NULL) {
		hPrevFont = SelectObject(hdc, hFont);
		DeleteObject(hPrevFont);
		return TRUE;
	} else {
		return FALSE;
	}
}

int GetNumDisplayRows(HDC hdc, long width, long height, int subtractStatusBarHeight) {
	switch(rotation) {
		case ROTATION_NORMAL:
		case ROTATION_INVERTED:
			return (int)floor((height - subtractStatusBarHeight) / GetFontHeight(hdc));
		case ROTATION_ROR:
		case ROTATION_ROL:
		default:
			return (int)floor((width - subtractStatusBarHeight) / GetFontHeight(hdc));
	}
}

int GetNumImagesX(HDC hdc, long width, long height, int subtractStatusBarHeight) {
	switch (rotation) {
		case ROTATION_NORMAL:
		case ROTATION_INVERTED:
			return (width - previewSpacer) / (previewWidth + previewSpacer);
		case ROTATION_ROR:
		case ROTATION_ROL:
		default:
			return (height - previewSpacer) / (previewWidth + previewSpacer);
	}
}

int GetNumImagesY(HDC hdc, long width, long height, int subtractStatusBarHeight) {
	switch (rotation) {
		case ROTATION_NORMAL:
		case ROTATION_INVERTED:
			return (height - previewSpacer - subtractStatusBarHeight) / (previewHeight + previewSpacer);
		case ROTATION_ROR:
		case ROTATION_ROL:
		default:
			return (width - previewSpacer - subtractStatusBarHeight) / (previewHeight + previewSpacer);
	}
}

BOOL fileExists(char *filename) {
	WIN32_FIND_DATA wfd;
	HANDLE hFind;
	BOOL retval;

	hFind = FindFirstFile(filename, &wfd);
	retval = hFind != INVALID_HANDLE_VALUE;
	FindClose(hFind);

	return retval;
}

int GetFontHeight(HDC hdc) {
	TEXTMETRIC tm;
	if (!GetTextMetrics(hdc, &tm))
		MessageBox(hwndMain, "GetTextMetrics failed!", "Error", MB_OK);
	return tm.tmHeight - tm.tmExternalLeading;
}

void rotateDIB(PNGD_P2DINFO *p2d){
	int oldBytesPerRow, newBytesPerRow;
	int newWidth, newHeight;
	int oldWidth, oldHeight;
	int x, y;
	BYTE *oldbits, *newbits;
	short int bytesPerPixel;

	bytesPerPixel = p2d->lpdib->biBitCount >> 3;
	oldbits = (BYTE *)p2d->bits;
	oldWidth = p2d->lpdib->biWidth;
	oldHeight = p2d->lpdib->biHeight;
	oldBytesPerRow = (((p2d->lpdib->biWidth * p2d->lpdib->biBitCount)+31)/32)*4;

	switch(rotation) {
		case ROTATION_INVERTED:
			newWidth = p2d->lpdib->biWidth;
			newHeight = p2d->lpdib->biHeight;
			newBytesPerRow = (((newWidth * p2d->lpdib->biBitCount)+31)/32)*4;

			newbits = (BYTE *)malloc(newBytesPerRow * newHeight);
			for (y=0; y<newHeight; y++)
				for (x=0; x<newWidth; x++)
					memcpy(
						&newbits[y * newBytesPerRow + x * bytesPerPixel],
						&oldbits[(newHeight-1 - y) * oldBytesPerRow + (newWidth-1 - x) * bytesPerPixel],
						bytesPerPixel
					);

			break;
		case ROTATION_ROR:
			newWidth = p2d->lpdib->biHeight;
			newHeight = p2d->lpdib->biWidth;
			newBytesPerRow = (((newWidth * p2d->lpdib->biBitCount)+31)/32)*4;

			newbits = (BYTE *)malloc(newBytesPerRow * newHeight);

			for (y=0; y<newHeight; y++)
				for (x=0; x<newWidth; x++)
					memcpy(
						&newbits[y * newBytesPerRow + x * bytesPerPixel],
						&oldbits[x * oldBytesPerRow + (newHeight-1 - y) * bytesPerPixel],
						bytesPerPixel
					);
			break;
		case ROTATION_ROL:
			newWidth = p2d->lpdib->biHeight;
			newHeight = p2d->lpdib->biWidth;
			newBytesPerRow = (((newWidth * p2d->lpdib->biBitCount)+31)/32)*4;

			newbits = (BYTE *)malloc(newBytesPerRow * newHeight);
			for (y=0; y<newHeight; y++)
				for (x=0; x<newWidth; x++)
					memcpy(
						&newbits[y * newBytesPerRow + x * bytesPerPixel],
						&oldbits[(oldHeight-1 - x) * oldBytesPerRow + y * bytesPerPixel],
						bytesPerPixel
					);

			break;
		default:
			// do nothing
			return;
	}

	free(oldbits);
	p2d->bits = newbits;
	p2d->lpdib->biWidth = newWidth;
	p2d->lpdib->biHeight = newHeight;
	p2d->lpdib->biSizeImage = newBytesPerRow * newHeight;
}

UINT getStringWidth(HDC hdc, const char *text) {
	SIZE size;

	if (!GetTextExtentPoint32(hdc, text, strlen(text), &size)) {
		MessageBox(hwndMain, "GetTextExtentPoint32 failed!", "Error", MB_OK);
		return 0;
	} else
		return size.cx;
}

BOOL drawImageToHDC(HDC hdc, HBITMAP hbitmap, GAME *game, RECT *drawArea);
BOOL DrawHighlightBoxToHDC(HDC hdc, int selectedGameIndex, RECT *drawArea, int fontHeight, int numDisplayRows);
void mapPoint(long *x, long *y, long width, long height);

BOOL DrawScreenImageViewOld(HDC hdc, HBITMAP hbitmap, RECT *rect) {
	BITMAP bmi;
	HBRUSH bgBrush, oldBrush;
	LOGBRUSH lb;
	HPEN selectPen, oldPen;
	RECT rectOLO, rectImg;
	int imagesX, imagesY;
	int x, y;
	int oldFI;
	int fontHeight;
	int sbHeight;
	char desc[1024];

	GetObject(hbitmap, sizeof(BITMAP), &bmi);

	switch (rotation) {
		case ROTATION_NORMAL:
			sbHeight = bmi.bmHeight - rect->bottom;
			break;
		case ROTATION_INVERTED:
			sbHeight = rect->top;
			break;
		case ROTATION_ROR:
			sbHeight = rect->left;
			break;
		case ROTATION_ROL:
			sbHeight = bmi.bmWidth - rect->right;
			break;
	}

	imagesX = GetNumImagesX(hdcTile, rect->right, rect->bottom, sbHeight);
	imagesY = GetNumImagesY(hdcTile, rect->right, rect->bottom, sbHeight);

	// Determine the first game index to be displayed
	oldFI = firstGameOnScreen;
	while (currentGameSelection < firstGameOnScreen)
		firstGameOnScreen = MAX(0, firstGameOnScreen - imagesX * imagesY);

	while (currentGameSelection > firstGameOnScreen + imagesX * imagesY-1)
		firstGameOnScreen = MIN(gameDisplayCount - imagesX * imagesY, firstGameOnScreen + imagesX * imagesY);

	// Redraw hdcTile if it's out of date
	if (firstGameOnScreen != oldFI || redrawBrowseMode) {
		redrawBrowseMode = FALSE;

		// clear the tile hdc
		bgBrush = CreateSolidBrush(previewBackColor);
		FillRect(hdcTile, rect, bgBrush);
		DeleteObject(bgBrush);

		for (y=0; y<imagesY; y++) {
			for (x=0; x<imagesX; x++) {
				if (firstGameOnScreen + y * imagesX + x < gameDisplayCount) {
					switch (rotation) {
						case ROTATION_NORMAL:
							rectImg.left = x % imagesX * (previewWidth + previewSpacer) + previewSpacer;
							rectImg.top = y % imagesY * (previewHeight + previewSpacer) + previewSpacer;
							rectImg.right = rectImg.left + previewWidth;
							rectImg.bottom = rectImg.top + previewHeight;
							break;
						case ROTATION_INVERTED:
							rectImg.left = rect->right - x % imagesX * (previewWidth + previewSpacer) - previewSpacer - previewWidth;
							rectImg.top = rect->bottom - y % imagesY * (previewHeight + previewSpacer) - previewSpacer - previewHeight;
							rectImg.right = rectImg.left + previewWidth;
							rectImg.bottom = rectImg.top + previewHeight;
							break;
						case ROTATION_ROR:
							rectImg.left = rect->right - y % imagesY * (previewHeight + previewSpacer) - previewSpacer - previewHeight;
							rectImg.top = x % imagesX * (previewWidth + previewSpacer) + previewSpacer;
							rectImg.right = rectImg.left + previewHeight;
							rectImg.bottom = rectImg.top + previewWidth;
							break;
						case ROTATION_ROL:
							rectImg.left = y % imagesY * (previewHeight + previewSpacer) + previewSpacer;
							rectImg.top = rect->bottom - x % imagesX * (previewWidth + previewSpacer) - previewSpacer - previewWidth;
							rectImg.right = rectImg.left + previewHeight;
							rectImg.bottom = rectImg.top + previewWidth;
							break;
					}
					drawImageToHDC(hdcTile, hbitmap, gameDisplayList[firstGameOnScreen + y * imagesX + x], &rectImg);
				}
			}
		}
	}

	// Copy the tile buffer to the offscreen buffer
	BitBlt(hdc, rect->left, rect->top, rect->right, rect->bottom, hdcTile, rect->left, rect->top, SRCCOPY);

	// Draw the highlight around the selected game
	rectOLO.left = (currentGameSelection - firstGameOnScreen) % imagesX * (previewWidth + previewSpacer) + previewSpacer / 2;
	rectOLO.top = (currentGameSelection - firstGameOnScreen) / imagesX * (previewHeight + previewSpacer) + previewSpacer / 2;
	rectOLO.right = rectOLO.left + previewWidth + previewSpacer;
	rectOLO.bottom = rectOLO.top + previewHeight + previewSpacer;

	// Map the points to the right rotation
	mapPoint(&rectOLO.left, &rectOLO.top, rect->right, rect->bottom);
	mapPoint(&rectOLO.right, &rectOLO.bottom, rect->right, rect->bottom);

	selectPen = CreatePen(PS_SOLID, previewSpacer, previewHighlightBackColor);
	oldPen = SelectObject(hdc, selectPen);
	lb.lbStyle = BS_HOLLOW;
	lb.lbColor = 0;
	lb.lbHatch = 0;
	bgBrush = CreateBrushIndirect(&lb);
	oldBrush = SelectObject(hdc, bgBrush);

	RoundRect(hdc, rectOLO.left, rectOLO.top, rectOLO.right, rectOLO.bottom, previewSpacer, previewSpacer);

	SelectObject(hdc, oldBrush);
	DeleteObject(bgBrush);
	SelectObject(hdc, oldPen);
	DeleteObject(selectPen);

	return TRUE;
}

BOOL drawImageToHDC(HDC hdc, HBITMAP hbitmap, GAME *game, RECT *drawArea) {
#if DEBUG
	LARGE_INTEGER start, end, freq;
#endif
	PNGD_P2DINFO p2d;
	char *filename;
	BOOL exists;
	float scaleFactor;
	char *Dir;

	exists=FALSE;

#if DEBUG
	QueryPerformanceFrequency((LARGE_INTEGER*)&freq);
	QueryPerformanceCounter((LARGE_INTEGER*)&start);
#endif

	filename = (char *)malloc(strlen(marqueeDir) + 14);	// allow room for "\8.3"
	sprintf(filename, "%s\\%s.png", marqueeDir, game->romname);
	if (fileExists(filename)) exists = TRUE;

	if(exists==FALSE)
	{
		if (game->cloneof) {
			sprintf(filename, "%s\\%s.png", marqueeDir, game->cloneof);
			if (fileExists(filename)) exists = TRUE;
		}
	}

	if(exists==FALSE)
	{
			sprintf(filename, "%s\\generic.png",marqueeDir);
			if (fileExists(filename)) exists = TRUE;
	}

	if (exists) {
		ZeroMemory((void*)&p2d,sizeof(PNGD_P2DINFO));

		p2d.structsize=sizeof(PNGD_P2DINFO);
		p2d.pngfn=filename;

		if(read_png_to_dib(&p2d)) {    // returns 0 on success
			return FALSE;
		}

		// rotate the picture if needed
		rotateDIB(&p2d);

		//scaleFactor = (float)(drawArea->right - drawArea->left) / p2d.lpdib->biWidth;
		//if (p2d.lpdib->biHeight * scaleFactor > drawArea->bottom - drawArea->top)
		//	scaleFactor = (float) (drawArea->bottom - drawArea->top) / p2d.lpdib->biHeight;

		// Do a custom stretch for marquees
		StretchDIBits(hdc,
			(int)drawArea->left,
			(int)drawArea->top,
			(int)drawArea->right - drawArea->left, (int)drawArea->bottom - drawArea->top,
			0, 0, p2d.lpdib->biWidth, p2d.lpdib->biHeight, p2d.bits, (LPBITMAPINFO)p2d.lpdib, DIB_RGB_COLORS, SRCCOPY);

		//StretchDIBits(hdc,
		//	(int)(drawArea->right - drawArea->left - p2d.lpdib->biWidth * scaleFactor) / 2 + drawArea->left,
		//	(int)(drawArea->bottom - drawArea->top - p2d.lpdib->biHeight * scaleFactor) / 2 + drawArea->top,
		//	(int)(p2d.lpdib->biWidth * scaleFactor), (int)(p2d.lpdib->biHeight * scaleFactor),
		//	0, 0, p2d.lpdib->biWidth, p2d.lpdib->biHeight, p2d.bits, (LPBITMAPINFO)p2d.lpdib, DIB_RGB_COLORS, SRCCOPY);

		GlobalFree(p2d.lpdib);

#if DEBUG
		QueryPerformanceCounter((LARGE_INTEGER*)&end);
		printf("Loaded image: [%s] in %.2fms\n", filename, (end.QuadPart - start.QuadPart) * 1000.0 / freq.QuadPart);
#endif
	}
	free(filename);

	return TRUE;
}

void mapPoint(long *x, long *y, long width, long height) {
	int tmp;
	switch (rotation) {
		case ROTATION_NORMAL:
			// do nothing
			break;
		case ROTATION_INVERTED:
			*x = width - *x;
			*y = height - *y;
			break;
		case ROTATION_ROR:
			tmp = *x;
			*x = width - *y;
			*y = tmp;
			break;
		case ROTATION_ROL:
			tmp = *y;
			*y = height - *x;
			*x = tmp;
			break;
	}
}

BOOL DrawScreenImageView(HDC hdc, HBITMAP hbitmap, RECT *rect) {
	BITMAP bmi;
	HBRUSH bgBrush, oldBrush;
	LOGBRUSH lb;
	HPEN selectPen, oldPen;
	RECT rectOLO, rectImg;
	int imagesX, imagesY;
	int x, y;
	int oldFI;
	int fontHeight;
	int sbHeight;
	char desc[1024];

	GetObject(hbitmap, sizeof(BITMAP), &bmi);

	switch (rotation) {
		case ROTATION_NORMAL:
			sbHeight = bmi.bmHeight - rect->bottom;
			break;
		case ROTATION_INVERTED:
			sbHeight = rect->top;
			break;
		case ROTATION_ROR:
			sbHeight = rect->left;
			break;
		case ROTATION_ROL:
			sbHeight = bmi.bmWidth - rect->right;
			break;
	}

	previewWidth = rect->right / 2;
	previewHeight = (rect->bottom - sbHeight) / 5;
	imagesX = 1;
	imagesY = GetNumImagesY(hdcTile, rect->right, rect->bottom, sbHeight) + 1;

	// Determine the first game index to be displayed
	oldFI = firstGameOnScreen;
	while (currentGameSelection < firstGameOnScreen)
		firstGameOnScreen = MAX(0, firstGameOnScreen - imagesX * imagesY);

	while (currentGameSelection > firstGameOnScreen + imagesX * imagesY-1)
		firstGameOnScreen = MIN(gameDisplayCount - imagesX * imagesY, firstGameOnScreen + imagesX * imagesY);

	// Redraw hdcTile if it's out of date
	if (firstGameOnScreen != oldFI || redrawBrowseMode) {
		redrawBrowseMode = FALSE;

		// clear the tile hdc
		bgBrush = CreateSolidBrush(previewBackColor);
		FillRect(hdcTile, rect, bgBrush);
		DeleteObject(bgBrush);

		for (y=0; y<imagesY; y++) {
			for(x=0; x<imagesX; x++) {
				if (firstGameOnScreen + y * imagesX + x < gameDisplayCount) {
					switch (rotation) {
						case ROTATION_NORMAL:
							rectImg.left = x % imagesX * (previewWidth + previewSpacer) + previewSpacer;
							rectImg.top = y % imagesY * (previewHeight + previewSpacer) + previewSpacer;
							rectImg.right = rectImg.left + previewWidth;
							rectImg.bottom = rectImg.top + previewHeight;
							break;
						case ROTATION_INVERTED:
							rectImg.left = rect->right - x % imagesX * (previewWidth + previewSpacer) - previewSpacer - previewWidth;
							rectImg.top = rect->bottom - y % imagesY * (previewHeight + previewSpacer) - previewSpacer - previewHeight;
							rectImg.right = rectImg.left + previewWidth;
							rectImg.bottom = rectImg.top + previewHeight;
							break;
						case ROTATION_ROR:
							rectImg.left = rect->right - y % imagesY * (previewHeight + previewSpacer) - previewSpacer - previewHeight;
							rectImg.top = x % imagesX * (previewWidth + previewSpacer) + previewSpacer;
							rectImg.right = rectImg.left + previewHeight;
							rectImg.bottom = rectImg.top + previewWidth;
							break;
						case ROTATION_ROL:
							rectImg.left = y % imagesY * (previewHeight + previewSpacer) + previewSpacer;
							rectImg.top = rect->bottom - x % imagesX * (previewWidth + previewSpacer) - previewSpacer - previewWidth;
							rectImg.right = rectImg.left + previewHeight;
							rectImg.bottom = rectImg.top + previewWidth;
							break;
					}
					drawImageToHDC(hdcTile, hbitmap, gameDisplayList[firstGameOnScreen + y * imagesX + x], &rectImg);
				}
			}
		}
	}

	// Copy the tile buffer to the offscreen buffer
	BitBlt(hdc, rect->left, rect->top, rect->right, rect->bottom, hdcTile, rect->left, rect->top, SRCCOPY);

	// Draw the highlight around the selected game
	rectOLO.left = (currentGameSelection - firstGameOnScreen) % imagesX * (previewWidth + previewSpacer) + previewSpacer / 2;
	rectOLO.top = (currentGameSelection - firstGameOnScreen) / imagesX * (previewHeight + previewSpacer) + previewSpacer / 2;
	rectOLO.right = rectOLO.left + previewWidth + previewSpacer;
	rectOLO.bottom = rectOLO.top + previewHeight + previewSpacer;

	// Map the points to the right rotation
	mapPoint(&rectOLO.left, &rectOLO.top, rect->right, rect->bottom);
	mapPoint(&rectOLO.right, &rectOLO.bottom, rect->right, rect->bottom);

	selectPen = CreatePen(PS_SOLID, previewSpacer, previewHighlightBackColor);
	oldPen = SelectObject(hdc, selectPen);
	lb.lbStyle = BS_HOLLOW;
	lb.lbColor = 0;
	lb.lbHatch = 0;
	bgBrush = CreateBrushIndirect(&lb);
	oldBrush = SelectObject(hdc, bgBrush);

	RoundRect(hdc, rectOLO.left, rectOLO.top, rectOLO.right, rectOLO.bottom, previewSpacer, previewSpacer);

	SelectObject(hdc, oldBrush);
	DeleteObject(bgBrush);
	SelectObject(hdc, oldPen);
	DeleteObject(selectPen);

	return TRUE;
}
