#ifndef HASH_H
#define HASH_H 

#include <windows.h>

// Linked list stuff
typedef struct _LLNODE {
	void *data;
	struct _LLNODE *next;
} LLNODE;

LLNODE *insertLLNodeAfter(void *data, LLNODE *previous);
void destroyNode(LLNODE *node, LLNODE *prev);
void destroyLL(LLNODE *firstNode);


// Hash table stuff
typedef struct _HASHELEMENT {
	void *key;
	void *value;
} HASHELEMENT;

typedef struct _HASHTABLE {
	LLNODE **buckets;		// hash buckets
	UINT numBuckets;		// number of buckets
	UINT (*hashFunction)(const void *item, const UINT numBuckets);	// get the hash value for an item
	BOOL (*equals)(void *item1, void *item2);	// determine if two items are equal
} HASHTABLE;

HASHTABLE *createHT(
	UINT numBuckets,
	UINT (*hashFunction)(const void *item, const UINT numBuckets),
	BOOL (*equals)(void *item1, void *item2)
);
void destroyHT(HASHTABLE *ht);
void addHTItem(void *key, void *value, HASHTABLE *ht);
void *removeHTItem(void *key, HASHTABLE *ht);
void *findHTItem(void *key, HASHTABLE *ht);

#endif
