#ifndef MAMEPARSER_H
#define MAMEPARSER_H

#include <windows.h>
#include "hash.h"

#define VERTICAL 1
#define HORIZONTAL 2

// All the available controller types (as of v0.53)
#define CONTROL_DIAL		1
#define CONTROL_DOUBLE_JOY8WAY	2
#define CONTROL_JOY4WAY		4
#define CONTROL_JOY8WAY		8
#define CONTROL_PADDLE		16
#define CONTROL_STICK		32
#define CONTROL_TRACKBALL	64

// Representation of a single rom
typedef struct _ROM {
	char *name;		// Name of the rom
	unsigned long size;	// Size of the rom
	unsigned long crc;	// CRC
} ROM;

// Video info for a rom
typedef struct _VIDEO {
	char *screen;		// Type of drawing
	short orientation;	// VERTICAL or HORIZONTAL
	short xres;		// X resolution
	short yres;		// Y resolution
} VIDEO;

// Input info for a rom
typedef struct _INPUT {
	short players;		// Number of players
	short buttons;		// Number of buttons
	short coins;		// Number of coin slots
	short control;		// Type of controller (choices are above)
} INPUT;

// Representation of a single game
typedef struct _GAME {
	char *romname;		// Name of the rom
	char *description;	// Description of the rom
	int year;		// Year the rom was made
	char *manufacturer;	// Who made it
	char *cloneof;		// Name of the parent rom (if there is one)
	char *romof;		// Uses files from this rom (if any)
	ROM **roms;		// Roms this one contains
	int romcount;		// Number of roms in this game
	VIDEO *video;		// Video information
	INPUT *input;		// Input information
	char *driverstatus;
	char *colorstatus;
	char *soundstatus;
	BOOL isResource;	// Is this a bios file
	BOOL exists;		// Does the rom exist on the filesystem
	char **mameInfoLines;	// Mameinfo.dat information
	int numMameInfoLines;	// Number of lines of mameinfo.dat information
	struct _GAME *parent;	// If the game is a clone, this will point to the parent
} GAME;

GAME **gameList;	// List of all games (don't modify)
int gameCount;		// Number of games in gameList (don't modify)
HASHTABLE *gameListHT;	// All games in a hashtable for easy access (don't modify)

GAME **gameDisplayList;	// List of games to display (pointers into gameList)
int gameDisplayCount;	// Number of games in gameDisplayList

BOOL loadListinfoFile(char *filename);
void freeGameList();
void freeGame(GAME *game);

#endif
