#include <windows.h>
#include "menu.h"
#include "drawing.h"
#include "launcher.h"
#include "input.h"

LRESULT CALLBACK MenuWndProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
HWND createMenuWindow(int x, int y, int width, int height);
void processKeyStrokesMenu(MENU_WINDOW *mw);
int runMessageLoopMenu(MENU_WINDOW *mw);
BOOL DrawMenu(MENU_WINDOW *mw);

MENU_WINDOW **allWindows;			// All menu windows that are open
int allWindowsCount = 0;			// Number of elements in allWindows

// Displays a menu that shows the given information and returns the index of the user's selection
// Returns FALSE on error, and TRUE otherwise
BOOL displayMenu(
	MENU_ITEM **items,			// Items to display
	int numItems				// Number of items
) {
	MENU_WINDOW *mw;
	int retval;
	int i;
	
	mw = (MENU_WINDOW *)malloc(sizeof(MENU_WINDOW));
	memset(mw, 0, sizeof(MENU_WINDOW));
	mw->items = items;
	mw->numItems = numItems;
	
	allWindows = (MENU_WINDOW **)realloc(allWindows, sizeof(MENU_WINDOW *) * ++allWindowsCount);
	allWindows[allWindowsCount - 1] = mw;
	
	mw->hwnd = createMenuWindow(0, 0, 100, 100);
	
	if (!mw->hwnd) {
		MessageBox(NULL, "Failed to create message window!", "Error!", MB_OK);
		free(mw);
		allWindows[allWindowsCount - 1] = NULL;
		return FALSE;
	}

	// Show the window
	ShowWindow(mw->hwnd, SW_SHOWDEFAULT);
	UpdateWindow(mw->hwnd);
	
	// Clear out any stray key presses
	for (i=0; i<255; i++) GetAsyncKeyState(i);
	
	runMessageLoopMenu(mw);
	
	DestroyWindow(mw->hwnd);
	retval = mw->selectedIndex;
	
	mw->hwnd = NULL;
	free(mw);
	mw = NULL;
	
	return TRUE;
}

// If not already registered, register the class for menu windows
BOOL registerMenuClass() {
	WNDCLASS wc;
	
	if (!GetClassInfo(hInst, "MenuWndClass", &wc)) {
		wc.style = CS_HREDRAW | CS_VREDRAW;
		wc.lpfnWndProc = (WNDPROC)MenuWndProc;
		wc.cbClsExtra = 0;
		wc.cbWndExtra = 0;
		wc.hInstance = hInst;
		wc.hIcon = LoadIcon(NULL, IDI_APPLICATION);
		wc.hCursor = LoadCursor(NULL, IDC_ARROW);
		wc.hbrBackground = GetStockObject(BLACK_BRUSH);
		wc.lpszMenuName =  "MenuMenu";
		wc.lpszClassName = "MenuWndClass";
		return RegisterClass(&wc);
	} else
		return TRUE;
}

// Create the window for the new menu
HWND createMenuWindow(int x, int y, int width, int height) {
	if (registerMenuClass())
		return CreateWindow(
			"MenuWndClass", NULL, WS_POPUP | WS_THICKFRAME,
			x, y, width, height,
			hwndMain, NULL, hInst, NULL);
	else
		return NULL;
}

int runMessageLoopMenu(MENU_WINDOW *mw) {
	LARGE_INTEGER start, end, freq;
	LONGLONG update_ticks_pc;
	MSG msg;
	
	// Get ticks-per-second of the performance counter.
	//   Note the necessary typecast to a LARGE_INTEGER structure
	if (!QueryPerformanceFrequency((LARGE_INTEGER*)&freq)) {
		MessageBox(NULL, "Hardware doesn't support QueryPerformanceFrequency()!", "Error", MB_OK);
		return -1;
	}
	
	// Convert milliseconds per move to performance counter units per move.
	update_ticks_pc = UPDATE_TICKS_MS * freq.QuadPart / 1000;
	
	// Initialize the counter.
	QueryPerformanceCounter((LARGE_INTEGER*)&start);
	
	// the main message loop begins here
	while (TRUE) {
		if (PeekMessage(&msg, NULL, 0, 0, PM_NOREMOVE)) {
			if (!GetMessage(&msg, NULL, 0, 0) || mw->closed) break;
  			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
		QueryPerformanceCounter((LARGE_INTEGER*)&end);
		
		if ((end.QuadPart - start.QuadPart) >= update_ticks_pc) {
			processKeyStrokesMenu(mw);
			start.QuadPart += update_ticks_pc;
		} else {
			Sleep(1);	// This sleeps for 10ms on my machine
		}
	} // End of message loop.
	
	return msg.wParam;
}

// Find the info for the given hwnd
MENU_WINDOW *findMenuWindow(HWND hwnd) {
	int i;
	
	for (i=0; i<allWindowsCount; i++)
		if (allWindows[i] && allWindows[i]->hwnd == hwnd)
			return allWindows[i];
	
	return NULL;
}

// Message processing function for the menu
LRESULT CALLBACK MenuWndProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam) {
	MENU_WINDOW *mw;
	
	if (uMsg == WM_NCCREATE) {	// First message sent, grab the hwnd out of it
		mw = allWindows[allWindowsCount - 1];
		mw->hwnd = hwnd;
	} else {
		mw = findMenuWindow(hwnd);
	}
	
	if (!mw)
		return 0;
	
	switch(uMsg) {
		case WM_CREATE:
			// get the main dc
			mw->onHDC = GetDC(hwnd);
			mw->offHDC = CreateCompatibleDC(mw->onHDC);
			break;
		case WM_DESTROY:
			if (mw->offHDC)
				DeleteDC(mw->offHDC);
			if (mw->onHDC)
				ReleaseDC(hwnd, mw->onHDC);
			if (mw->offHBitmap)
				DeleteObject(mw->offHBitmap);
			break;
		case WM_SIZE:
			{
				RECT rect, rectW, rectS;
				long width;
				long height;
				long descW, optW;
				long correctW, correctH;
				int i, j;
				HBITMAP hOldBitmap;
				int x, y;
				
				// get the width & height of the window
				GetClientRect(hwnd, &rect);
				width = rect.right - rect.left;
				height = rect.bottom - rect.top;
				
				// create the offscreen dc & bitmap
				mw->offHBitmap = CreateCompatibleBitmap(mw->onHDC, width, height);
				hOldBitmap = (HBITMAP)SelectObject(mw->offHDC, mw->offHBitmap);
				DeleteObject(hOldBitmap);
				
				SetFont(mw->offHDC, fontFace, menuFontSize, rotation, FALSE);
				
				// Calculate the width and height needed to display everything
				descW = optW = 0;
				for (i=0; i < mw->numItems; i++) {
					for (j=0; j < mw->items[i]->numOptions; j++) {
						optW = MAX((UINT)optW, getStringWidth(mw->offHDC, mw->items[i]->options[j]));
					}
					descW = MAX((UINT)descW, getStringWidth(mw->offHDC, mw->items[i]->description));
				}
				switch (rotation) {
					case ROTATION_NORMAL:
					case ROTATION_INVERTED:
						correctW = descW + optW + MENU_PAD + LEFT_PAD + LEFT_PAD;
						correctH = GetFontHeight(mw->offHDC) * mw->numItems;
						break;
					case ROTATION_ROR:
					case ROTATION_ROL:
						correctW = GetFontHeight(mw->offHDC) * mw->numItems;
						correctH = descW + optW + MENU_PAD + LEFT_PAD + LEFT_PAD;
						break;
				}
				if (correctW != width || correctH != height) {
					GetWindowRect(hwnd, &rectW);
					GetWindowRect(hwndMain, &rectS);
					x = MAX(0, (rectS.right - rectS.left - correctW) / 2);
					y = MAX(0, (rectS.bottom - rectS.top - correctH) / 2);
					MoveWindow(hwnd, x, y, correctW + rectW.right - rectW.left - rect.right, correctH + rectW.bottom - rectW.top - rect.bottom, TRUE);
				}
				
				DrawMenu(mw);
				InvalidateRect(hwnd, NULL, FALSE);
				break;
			}
		case WM_PAINT:
			{
				PAINTSTRUCT paintStruct;
				RECT rect;
				BeginPaint(hwnd, &paintStruct);
	
				GetClientRect(hwnd, &rect);
				
				if(mw->offHDC)
					BitBlt(mw->onHDC, rect.left, rect.top, rect.right, rect.bottom, mw->offHDC, 0, 0, SRCCOPY);
				
				EndPaint(hwnd, &paintStruct);
				break;
			}
		case WM_SETFOCUS:
			inputEnabled = TRUE;
			break;
		case WM_KILLFOCUS:
			inputEnabled = FALSE;
			break;
		default:
			return(DefWindowProc(hwnd, uMsg, wParam, lParam));
	}
	return 0;
}

BOOL DrawMenu(MENU_WINDOW *mw) {
	int numDisplayRows;
	int width, height;
	int i;
	int fontHeight;
	int x, y;
	char *description;
	MENU_ITEM *item;
	RECT rect;
	int textWidth;
	HBRUSH bgBrush;
	
	// get the width & height of the window
	GetClientRect(mw->hwnd, &rect);
	width = rect.right - rect.left;
	height = rect.bottom - rect.top;
	
	SetFont(mw->offHDC, fontFace, menuFontSize, rotation, FALSE);
	SetTextColor(mw->offHDC, foreColor);
	SetBkMode(mw->offHDC, TRANSPARENT);
	
	// clear the screen (to the background color
	bgBrush = CreateSolidBrush(backColor);
	FillRect(mw->offHDC, &rect, bgBrush);
	DeleteObject(bgBrush);
	
	numDisplayRows = GetNumDisplayRows(mw->offHDC, width, height, 0);
	fontHeight = GetFontHeight(mw->offHDC);
	
	if (mw->selectedIndex < mw->firstIndex || mw->selectedIndex > mw->firstIndex + numDisplayRows-1)
		mw->firstIndex = MIN(
			MAX(mw->selectedIndex - numDisplayRows / 2, 0),
			mw->numItems - 1 - numDisplayRows
		);
	
	for (i = 0; i < numDisplayRows && i + mw->firstIndex < mw->numItems; i++) {
		item = mw->items[i + mw->firstIndex];
		
		// Text on the left side
		description = item->description;
		switch (rotation) {
			case ROTATION_NORMAL:
				x = LEFT_PAD;
				y = i * fontHeight;
				break;
			case ROTATION_INVERTED:
				x = width - LEFT_PAD;
				y = height - i * fontHeight;
				break;
			case ROTATION_ROR:
				x = width - i * fontHeight;
				y = LEFT_PAD;
				break;
			case ROTATION_ROL:
				x = i * fontHeight;
				y = height - LEFT_PAD;
				break;
		}
		// bold the font when drawing the selected game
		if (mw->selectedIndex == (mw->firstIndex + i))
			SetFont(mw->offHDC, fontFace, menuFontSize, rotation, TRUE);
		
		TextOut(mw->offHDC, x, y, description, strlen(description));
		
		// Text on the right side
		if (item->numOptions > 0 || item->numSubMenuItems > 0) {
			if (item->numSubMenuItems == 0)
				description = item->options[item->selectedOption];
			else
				description = "[...]";
			
			textWidth = getStringWidth(mw->offHDC, description);
			
			switch (rotation) {
				case ROTATION_NORMAL:
					x = width - LEFT_PAD - textWidth;
					y = i * fontHeight;
					break;
				case ROTATION_INVERTED:
					x = LEFT_PAD + textWidth;
					y = height - i * fontHeight;
					break;
				case ROTATION_ROR:
					x = width - i * fontHeight;
					y = height - LEFT_PAD - textWidth;
					break;
				case ROTATION_ROL:
					x = i * fontHeight;
					y = LEFT_PAD + textWidth;
					break;
			}
			TextOut(mw->offHDC, x, y, description, strlen(description));
		}

		if (mw->selectedIndex == (mw->firstIndex + i))
			SetFont(mw->offHDC, fontFace, menuFontSize, rotation, FALSE);
	}

	return TRUE;
}

void menuCleanup() {
	int i;
	for (i=0; i<allWindowsCount; i++)
		if (allWindows[i]) {
			DestroyWindow(allWindows[i]->hwnd);
			free (allWindows[i]);
			allWindows[i] = NULL;
		}
	free(allWindows);
	allWindows = NULL;
	allWindowsCount = 0;
	UnregisterClass("MenuWndClass", hInst);
}

void processKeyStrokesMenu(MENU_WINDOW *mw) {
	int i, newRotation;
	BOOL p1Up, p1Down, p1Pgup, p1Pgdown, p1Exit;
	BOOL p2Up, p2Down, p2Pgup, p2Pgdown, p2Exit;
	
	// Record the states of all the keys
	p1Up = keyPressed(keySelectUp) || joyInPosition(joySelectUp);
	p1Down = keyPressed(keySelectDown) || joyInPosition(joySelectDown);
	p1Pgup = keyPressed(keySelectPageDown) || joyInPosition(joySelectPageDown);
	p1Pgdown = keyPressed(keySelectPageUp) || joyInPosition(joySelectPageUp);
	p1Exit = keyPressed(keyExit) || joyInPosition(joyExit);
	
	p2Up = keyPressed(keyP2SelectUp);
	p2Down = keyPressed(keyP2SelectDown);
	p2Pgup = keyPressed(keyP2SelectPageDown);
	p2Pgdown = keyPressed(keyP2SelectPageUp);
	p2Exit = keyPressed(keyP2Exit);
	
	// Flip the screen if requested
	if (invertPlayer2Screen) {
		switch (originalRotation) {
			case ROTATION_NORMAL:
				if (p2Up ||  p2Down || p2Pgup || p2Pgdown || p2Exit)
					newRotation = ROTATION_INVERTED;
				if (p1Up || p1Down || p1Pgup || p1Pgdown || p1Exit)
					newRotation = ROTATION_NORMAL;
				break;
			case ROTATION_INVERTED:
				if (p2Up ||  p2Down || p2Pgup || p2Pgdown || p2Exit)
					newRotation = ROTATION_NORMAL;
				if (p1Up || p1Down || p1Pgup || p1Pgdown || p1Exit)
					newRotation = ROTATION_INVERTED;
				break;
			case ROTATION_ROR:
				if (p2Up ||  p2Down || p2Pgup || p2Pgdown || p2Exit)
					newRotation = ROTATION_ROL;
				if (p1Up || p1Down || p1Pgup || p1Pgdown || p1Exit)
					newRotation = ROTATION_ROR;
				break;
			case ROTATION_ROL:
				if (p2Up ||  p2Down || p2Pgup || p2Pgdown || p2Exit)
					newRotation = ROTATION_ROR;
				if (p1Up || p1Down || p1Pgup || p1Pgdown || p1Exit)
					newRotation = ROTATION_ROL;
				break;
		}
		if (rotation != newRotation) {
			rotation = newRotation;
			redrawBrowseMode = TRUE;
		}
	}
	
	if (p1Up || p2Up) {			// Select next lower # item
		if (mw->selectedIndex > 0)
			mw->selectedIndex--;
		else
			mw->selectedIndex = mw->numItems - 1;
		DrawMenu(mw);
		InvalidateRect(mw->hwnd, NULL, FALSE);
	}
  
	if (p1Down || p2Down) {		// Select next higher # item
		if (mw->selectedIndex < mw->numItems-1)
			mw->selectedIndex++;
		else
			mw->selectedIndex = 0;
		DrawMenu(mw);
		InvalidateRect(mw->hwnd, NULL, FALSE);
	}
	
	if (p1Pgup || p2Pgup) {	// Select next option
		MENU_ITEM *mi = mw->items[mw->selectedIndex];
		if (mi->numSubMenuItems == 0) {
			if (mi->selectedOption < mi->numOptions-1)
				mi->selectedOption++;
			else
				mi->selectedOption = 0;
			DrawMenu(mw);
			InvalidateRect(mw->hwnd, NULL, FALSE);
		} else {
			RECT rect;
			GetWindowRect(mw->hwnd, &rect);
			displayMenu(mi->subMenu, mi->numSubMenuItems);
			
			// Clear out any stray key presses
			for (i=0; i<255; i++) GetAsyncKeyState(i);
		}
	}
	
	if (p1Pgdown || p2Pgdown) {		// Select previous option
		{
			MENU_ITEM *mi = mw->items[mw->selectedIndex];
			if (mi->selectedOption > 0)
				mi->selectedOption--;
			else
				mi->selectedOption = mi->numOptions-1;
			DrawMenu(mw);
			InvalidateRect(mw->hwnd, NULL, FALSE);
		}
	}
	
	if (p1Exit || p2Exit) {		// Exit the menu
		mw->closed = TRUE;
	}
}

